#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <Inventor/SoInput.h>
#include <Inventor/SoOutput.h>

#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>

#include <Inventor/actions/SoWriteAction.h>

#include <Inventor/draggers/SoTranslate1Dragger.h>

#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoLOD.h>
#include <Inventor/SoInput.h>
#include <Inventor/nodes/SoTriangleStripSet.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoText2.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoShapeHints.h>

#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/nodes/SoFont.h>

////////////////////////////////////////////////////////////
//
// Scene graph for a water molecule, displaying a 
// (reddish) oxygen atom, a hydrogen, and another hydrogen.
//
//                              root
//                                |
//                             WaterSep
//                                |
//         ---------------------------------------------
//         |                      |                    |
//      oxygenSep              hydrogen1Sep       hydrogen2Sep
//         |                      |                    |
//   -------------         -----------------          ----------------------
//   |           |         |        |       |         |        |           |
// material   sphere           translation sphere           translation sphere
//                                             ^                           |
//                                             |_. . . . . . . . . . . . ._.

int NumMolecules;
SoSeparator *root;
SoSeparator *windowSep;
SoSeparator *simulationSep;
SoSeparator *raySep;
SoText2 *windowText;
SoXtExaminerViewer *myViewer;
SoCamera *myCamera;
SoMFVec3f *glyVert;
SbVec3f *startingPos;
SbVec3f *offsets;
SoRotationXYZ *axisRot;
SoTransform *windowTrans2;
const double glyphRange = 75.0;
extern "C" void md_loop();
extern "C" void set_temperature_value(float);
extern "C" void beginAverage();
extern "C" void updateText(char *s);

////////////////////////////////////////////////////////////////////////
//
// ??
//

void 
do_md(void *data, SoSensor *){
   md_loop();
}

////////////////////////////////////////////////////////////////////////
//
// Updates onscreen message
//

void updateText(char *s){
   windowText->string = SbString(s);
}


////////////////////////////////////////////////////////////////////////
// Start average velocity calculations from the current position
// (aka reset startingPos[])
// Be sure to call AFTER animate() has been called.

void beginAverage(){
   int i;

   for (i=0;i<10;i++)
      printf("RESETTING AVERAGE START POSITIONS\n");
  
   SoSeparator *oxygenSep;
   SoTransform *oxygenTransform;
   SoSeparator *waterSep;
   SbVec3f oxygenTranslation;

   SbVec3f oldOxygen, newOxygen;
   SbVec3f oxygenMotion = SbVec3f(0.0, 0.0, 0.0);
   for (i = 0; i < NumMolecules; i++){
      waterSep = (SoSeparator *) (simulationSep->getChild(i) );
      oxygenSep = (SoSeparator *) (waterSep->getChild(0) );
      oxygenTransform = (SoTransform *) (oxygenSep->getChild(1) );

      startingPos[i] = oxygenTransform->translation.getValue();
   }

   for (i=0;i<NumMolecules;i++)
      offsets[i].setValue(0.0, 0.0, 0.0);
  
}

////////////////////////////////////////////////////////////////////////
//
// Keyboard events handles key presses
//

void keyboardEvents(void *data, SoEventCallback *eventCB){
   const SoEvent *ev = eventCB->getEvent();
   if (SO_KEY_PRESS_EVENT(ev, A))
      beginAverage();
}


////////////////////////////////////////////////////////////
//
// Update the molecules position, rays, glyph
//

void
animate(float **positions, int rays, float **raypos, float cellsize){
   static int Count = 0;

   int i;
   SoSeparator *oxygenSep;
   SoSeparator *hydrogen1Sep;
   SoSeparator *hydrogen2Sep;
   SoTransform *oxygenTransform;
   SoTransform *hydrogen1Transform;
   SoTransform *hydrogen2Transform;
   SoSeparator *waterSep;
   SbVec3f oxygenTranslation;
   SbVec3f hydrogen1Translation;
   SbVec3f hydrogen2Translation;

   SbVec3f oldPos;
   SbVec3f oxygenMotion = SbVec3f(0.0, 0.0, 0.0);
   myCamera = myViewer->getCamera();

   // rotate glyph icon
   axisRot->angle.setValue(axisRot->angle.getValue()+0.1);
  
   // orient glyph icon
   SbRotation invCamRot = myCamera->orientation.getValue();
   windowTrans2->rotation.setValue(invCamRot);
  
   // If first frame, reset starting position
   if (Count == 0){
      startingPos = new SbVec3f[NumMolecules];
      offsets = new SbVec3f[NumMolecules];
      for (i=0;i<NumMolecules;i++)
	 offsets[i].setValue(0.0, 0.0, 0.0);
      beginAverage();
   }
   else if (Count == 1){
      beginAverage();
      myViewer->viewAll();
      myCamera->pointAt(SbVec3f(0., 0., 0.));
   }


   // reset glyph vertex position
   SbVec3f *verts = glyVert->startEditing();
   for (int j=0; j<glyVert->getNum(); j++){
      verts[j].normalize();
   }
   glyVert->finishEditing();

   // Contribute each molecules average velocity to glyph vertices
   // Loop over each molecule
   for (i = 0; i < NumMolecules; i++){
      waterSep = (SoSeparator *) (simulationSep->getChild(i) );
      oxygenSep = (SoSeparator *) (waterSep->getChild(0) );
      hydrogen1Sep = (SoSeparator *) (waterSep->getChild(1) );
      hydrogen2Sep = (SoSeparator *) (waterSep->getChild(2) );
      oxygenTransform = (SoTransform *) (oxygenSep->getChild(1) );
      hydrogen1Transform = (SoTransform *) (hydrogen1Sep->getChild(0) );
      hydrogen2Transform = (SoTransform *) (hydrogen2Sep->getChild(0) );

      // save old position
      oldPos = oxygenTransform->translation.getValue();

      // set new positions
      oxygenTranslation.setValue(positions[3*i][0], positions[3*i][1],
				 positions[3*i][2]);
      oxygenTransform->translation = oxygenTranslation;
      hydrogen1Translation.setValue(positions[(3*i)+1][0], positions[(3*i)+1][1],
				    positions[(3*i)+1][2]);
      hydrogen1Transform->translation = hydrogen1Translation;
      hydrogen2Translation.setValue(positions[(3*i)+2][0], positions[(3*i)+2][1],
				    positions[(3*i)+2][2]);
      hydrogen2Transform->translation = hydrogen2Translation;

      // set offset - need cell size
      // this is for particles that wrap around
      for (int j=0; j<3; j++){
	 double diff = oxygenTranslation[j]-oldPos[j];
	 if (diff>cellsize/2.0)
	    offsets[i][j] -= cellsize;
	 else if (diff<-cellsize/2.0)
	    offsets[i][j] += cellsize;
      }
	
      // Set glyph vertex positions
      verts = glyVert->startEditing();
      for (int j=0; j<glyVert->getNum();j++){
	 SbVec3f radial = verts[j];
	 radial.normalize();
	 SbVec3f motion = oxygenTranslation+offsets[i] - startingPos[i];
	 double component = radial.dot(motion);
	 if (component<0.0)
	    //component = -component;
	    continue;
	 motion.normalize();
	 component = pow(radial.dot(motion), 6.0) *     // phong shading cos^6
	    component;                                  // times dot product
	 verts[j] += radial*component*10.0;
      }
      glyVert->finishEditing();
	
   }

   // scale glyph vertex positions
   // find max
   double max = 1.0,
      length;
   verts = glyVert->startEditing();
   for (int j=0; j<glyVert->getNum(); j++){
      length = verts[j].length();
      if (length>max)
	 max = length;
   }
   glyVert->finishEditing();
   // scale
   max = 2.0*cellsize/2.0*sqrt(2.0) / max; // scale factor (desired/max)
   verts = glyVert->startEditing();
   for (int j=0; j<glyVert->getNum(); j++){
      verts[j] *= max;
   }
   glyVert->finishEditing();

   // draw rays (bars)
   int ray;
   SoSeparator *raySubSep;
   SoTransform *rayXform;
   SoTranslation *rayTrans;
   SoCylinder *rayCyl;
   SbRotation rot;
   SbVec3f s, f;
   raySep->removeAllChildren();

   // Loop over each ray
   for (ray=0; ray<rays; ray++){
      s[0] = raypos[ray][0];
      s[1] = raypos[ray][1];
      s[2] = raypos[ray][2];
      f[0] = raypos[ray][3];
      f[1] = raypos[ray][4];
      f[2] = raypos[ray][5];

      // cylinder 
      rayCyl = new SoCylinder;
      rayCyl->height = (s-f).length();
      rayCyl->radius.setValue(0.1);

      // translation 
      rayTrans = new SoTranslation;
      rayTrans->translation.setValue(
	 SbVec3f(0.0, rayCyl->height.getValue()/2.0, 0.0) );

      // transform
      rayXform = new SoTransform();
      rot.setValue(SbVec3f(0.0, 1.0, 0.0), f-s);
      rayXform->rotation.setValue(rot);
      rayXform->translation.setValue(s);

      // sep
      raySubSep = new SoSeparator();
      raySubSep->addChild(rayXform);
      raySubSep->addChild(rayTrans);
      raySubSep->addChild(rayCyl);

      raySep->addChild(raySubSep);
   }

       
   // Write scene file for movie (disabled)
   if (0 && (Count >= 100)){
      char fileName[80];
     
      if ((Count % 5) == 0){
	 sprintf(fileName, "movie/waters.%05d.iv", Count);
	
	 SoOutput *output = new SoOutput();
	 output->openFile(fileName);
	 output->setBinary(TRUE);

	 SoWriteAction *writeAction = new SoWriteAction(output);
	 writeAction->apply(root);
	 output->closeFile();
      }

      // Quit after 4000 frames
      if (Count == 4000)
	 exit(0);
   }

   Count++;
}

void
updateMotionMagnitude(void *data, SoDragger *dragger){
   //float exponentScale = 0.5;

   SbVec3f trans = ((SoTranslate1Dragger *) dragger)->translation.getValue();
   //MotionMagnitude = MotionMagnitudeScale * exp(exponentScale * trans[0]);
   printf("\n dragger %f", trans[0]); 
   set_temperature_value(300.+200.*trans[0]);
}

/////////////////////////////////////////////////////////////////
//
// main

void
initialize(int nmols, float cellsize){
   NumMolecules = nmols;
   FILE *infile;
   SoSeparator *waterSep;
   float x, y, z;
   int i;

   // Initialize Inventor and Xt
   Widget myWindow = SoXt::init("Water");
   if (myWindow == NULL) exit(1);

   //////////////////////////////////////////////////////////////////////// 
   // Create different branches of root
   //
   //                               root
   //     .-----------------.---------^--------------------.
   //  callback        windowSep                          lod
   //             .------^-------.                      .----^------.
   //          transform       text               simulationSep   glyphSep
   //

   // root
   root  = new SoSeparator();

   // callback
   SoEventCallback *cb = new SoEventCallback();
   cb->addEventCallback(SoKeyboardEvent::getClassTypeId(),
			keyboardEvents, NULL);
   root->addChild(cb);

   // window sep
   windowSep = new SoSeparator();
   root->addChild(windowSep);
   SoSeparator *camSep = new SoSeparator();
   root->addChild(camSep);
   SoPerspectiveCamera *myPCamera = new SoPerspectiveCamera();
   camSep->addChild(myPCamera);
  
   // Lod
   SoLOD *lod = new SoLOD();
   simulationSep = new SoSeparator();
   SoSeparator *glyphSep = new SoSeparator();
   lod->addChild(simulationSep);
   lod->addChild(glyphSep);
   lod->range.set1Value(0, glyphRange);
   camSep->addChild(lod);

   // Set up windowSep

   // move to bottom corner
   SoTransform *windowTrans = new SoTransform();
   windowTrans->translation.setValue(-0.95, -0.95, 0.0);
   windowSep->addChild(windowTrans);

   // change font
   SoFont *myFont = new SoFont();
   myFont->name.setValue("Helvetica");
   myFont->size.setValue(12);
   //windowSep->addChild(myFont);

   // change color
   SoMaterial *textMat = new SoMaterial();
   textMat->diffuseColor.setValue(SbColor(1.0, 0.5, 0.5));
   windowSep->addChild(textMat);

   // add text
   windowText = new SoText2();
   //windowText->spacing.setValue(30.0);
   //windowText->justification.setValue(SoText2::LEFT);
   windowSep->addChild(windowText);

   // add light
   SoPointLight *myLight = new SoPointLight();
   myLight->intensity.setValue(1.0);
   myLight->location.setValue(SbVec3f(0.0, 0.0, 5.0));
   windowSep->addChild(myLight);

   // shift up a little
   SoTransform *windGlyTrans = new SoTransform();
   windGlyTrans->translation.setValue(0.1, 0.1, 0.2);
   windowSep->addChild(windGlyTrans);
  
   // add spinning rotation
   axisRot = new SoRotationXYZ();
   axisRot->axis.setValue(SoRotationXYZ::Z);
   //axisRot->angle.setValue(M_PI/2.0);
   //windowSep->addChild(axisRot);

   // add camera alignment of scene
   windowTrans2 = new SoTransform();
   windowTrans2->scaleFactor.setValue(1.0/cellsize/10.0,
				      1.0/cellsize/10.0,
				      1.0/cellsize/10.0);
   windowSep->addChild(windowTrans2);

   // add glyph
   windowSep->addChild(glyphSep);

   // create oxygen
   SoSphere *oxygen = new SoSphere();
   oxygen->radius.setValue(0.73);
   oxygen->ref();

   // create hydrogen
   SoSphere *hydrogen = new SoSphere();
   hydrogen->radius.setValue(0.32);
   hydrogen->ref();

   for(i=0; i<NumMolecules; i++){

      ////////////////////////////////////////////////////////////
      //
      //  Water molecule

      waterSep = new SoSeparator();


      ////////////////////////////////////////////////////////////
      //
      //  Oxygen

      SoSeparator *oxygenSep = new SoSeparator();
      SoMaterial *oxygenMaterial = new SoMaterial();
      oxygenMaterial->diffuseColor.setValue(0.8, 0.3, 0.3);
      SoTransform *oxygenTransform = new SoTransform();
      oxygenSep->addChild(oxygenMaterial);
      oxygenSep->addChild(oxygenTransform);
      oxygenSep->addChild(oxygen);

      ////////////////////////////////////////////////////////////
      //
      //  Hydrogen1

      SoSeparator *hydrogen1Sep = new SoSeparator();
      SoTransform *hydrogen1Transform = new SoTransform();
      hydrogen1Sep->addChild(hydrogen1Transform);
      hydrogen1Sep->addChild(hydrogen);

      ////////////////////////////////////////////////////////////
      //
      //  Hydrogen2

      SoSeparator *hydrogen2Sep = new SoSeparator();
      SoTransform *hydrogen2Transform = new SoTransform();
      hydrogen2Sep->addChild(hydrogen2Transform);
      hydrogen2Sep->addChild(hydrogen);


      waterSep->addChild(oxygenSep);
      waterSep->addChild(hydrogen1Sep);
      waterSep->addChild(hydrogen2Sep);

      simulationSep->addChild(waterSep);
   }

   // ray separator
   raySep = new SoSeparator();
   SoMaterial *rayMat = new SoMaterial();
   SbColor rayColor;
   rayColor.setValue(0.32, 0.70, 1.0);
   rayMat->ambientColor.setValue(rayColor*0.60);
   rayMat->diffuseColor.setValue(rayColor*0.52);
   rayMat->specularColor.setValue(rayColor*0.0);
   rayMat->emissiveColor.setValue(rayColor*0.18);
   rayMat->shininess.setValue(0.0);
   rayMat->transparency.setValue(0.0);
   simulationSep->addChild(rayMat);
   simulationSep->addChild(raySep);

   SoTimerSensor *sensor = new SoTimerSensor(do_md, root);
   //sensor->setInterval(SbTime(0.02)); // 50 frames per second
   sensor->setInterval(SbTime(0.01)); 
   sensor->schedule();

   // Add glyph sphere
   SoInput glyphInput;
   SoSeparator *glyphSphereSep;
   SoMaterial *glyphMat = new SoMaterial();
   glyphMat->diffuseColor.setValue(SbColor(0.9, 1.0, 0.9));
   glyphSep->addChild(glyphMat);
   glyphInput.openFile("sphere.16.32.iv");
   glyphSphereSep = SoDB::readAll(&glyphInput);
   glyphSep->addChild(glyphSphereSep);

   // find vertex node
   SoTriangleStripSet *glyScnTri = (SoTriangleStripSet *) glyphSphereSep->getChild(2);
   SoVertexProperty *glyVertProp = (SoVertexProperty*) glyScnTri->vertexProperty.getValue();
   glyVert = &(glyVertProp->vertex);

   // Insert dragger
   float draggerSizeWithinScene = 2.5;
   float translationScale = 1.;

   SoSeparator *heatDraggerSep = new SoSeparator;
   SoTransform *heatDraggerTransform = new SoTransform();
   heatDraggerTransform->translation.setValue(0.0, -10., 0.0);
   heatDraggerTransform->scaleFactor = SbVec3f(translationScale, translationScale,
					       translationScale) * draggerSizeWithinScene;
   heatDraggerSep->addChild(heatDraggerTransform);

   // Hand over the dragger subtree to the callback function
   // updateMotionMagnitude(), which will extract the current
   // value of the dragger's translation field.
   //
   // No reason to pass heatDraggerSep -- it's just in there
   // for "future use".

   SoTranslate1Dragger *heatDragger = new SoTranslate1Dragger();
   heatDragger->addMotionCallback (updateMotionMagnitude, heatDraggerSep);
   heatDraggerSep->addChild(heatDragger);
   //root->addChild(heatDraggerSep);

   myViewer = new SoXtExaminerViewer(myWindow);
   myViewer->setSceneGraph(root);
   myViewer->setTitle("Water");
   myViewer->show();
   myCamera = myViewer->getCamera();

   SoXt::show(myWindow);
   SoXt::mainLoop();
}
